/*
|	FILE:	RAST2SGM.C
|
|	This program converts the rast format output from RAST.EXE
|	back into SGML format. The real reason for this program is to allow
|	normalization of SGML files that for one reason or another
|	were generated without the aid of an SGML editor and are therefore
|	minimized. My SGML interpreter requires end tags.
|
|	RAST2SGM along with SGMLS and RAST form a processing stream that
|	can automatically normalize an SGML input file.
|
|	This version is a first cut to prove the concept. Attributes are
|	NOT handled yet!
|
|	Date:
|	10/28/93	Initial coding and test.
|	11/02/93	Drop #RE lines and optionally drop empty tags.
|	12/15/93	Added -w flag to wrap/nowrap on #RE
*/
#include <stdio.h>
#include <string.h>

#define FOREVER		for(;;)
#define START_CHAR	'['
#define END_CHAR	']'
#define DATA_CHAR	'|'
#define END_TAG		'/'
#define TRUE		1
#define FALSE		0
#define ALL		1
#define NONE		0
#define END		2

FILE * fd_in;
FILE * fd_out;

long cont_flag;
long empty_tag     = ALL;		/* Keep empty tags */
long wrap_lines    = TRUE;

char ptitle[]      = "RAST2SGM V1.3";
char pdate[]       = "November 1993";
char prev_tag[128] = "";

char file_in[128];
char file_out[128];

char in_buff[8192];
char out_buff[8192];

/***********************+-----------+
|			| init_rast |
|			+-----------+
*/
void init_rast()
{
	printf("\n%s\n\n",ptitle);
	cont_flag  = FALSE;

	if ((fd_in = fopen(file_in, "r")) == NULL){
		fprintf(stderr, "Cannot open input file.\n");
		exit(1);
	}

	if ((fd_out = fopen(file_out, "w")) == NULL){
		fprintf(stderr, "Cannot open output file.\n");
		exit(1);
	}
	switch(empty_tag){
	case NONE:
		printf("   Deleting empty tags\n");
		break;
	case ALL:
		printf("   Keeping empty tags\n");
		break;
	case END:
		printf("   Deleting end tags from empty tag sets\n");
		break;
	}
}
/***********************+-----------+
|			| exit_rast |
|			+-----------+
*/
void exit_rast()
{
	fclose(fd_in);
	fclose(fd_out);
}
/***********************+----------------+
|			| read_next_line |
|			+----------------+
*/
char* read_next_line()
{
	int  len;
	char *status;

	status = fgets(in_buff,256,fd_in);
		/*
		| get rid of trailing \n
		*/
	len    = strlen(in_buff);
	in_buff[len] = '\0';

	return(status);
}
/***********************+----------------+
|			| flush_out_buff |
|			+----------------+
*/
void flush_out_buff()
{
	strcat(out_buff,"\n");
	fputs(out_buff,fd_out);
// printf("%s",out_buff);
	out_buff[0] = '\0';
}
/***********************+-----------------+
|			| process_end_tag |
|			+-----------------+
| There should be data in the output buffer. Tack the end
| tag onto the output buffer, then flush it.
*/
void process_end_tag()
{
	int  i = 2;  
	int  j = 2;
	int  data_in_out_buff;
	int  status;

	char ch;
	char temp[80];
		/*
		| Start building the new end tag in temp string
		*/
	strcpy(temp,"</");
	ch        = in_buff[i++];
	do{
		temp[j++] = ch;
		ch        = in_buff[i++];
	}while(ch != END_CHAR);
	temp[j++] = '>';
	temp[j]   = '\0';
		/*
		| Decide what to do with the end tag...
		*/
	switch(empty_tag){
	case(ALL):		/* keep empty tags */
			/*
			| The temp string, which now contains the converted
			| end tag is tacked onto the output buffer...
			*/
		strcat(out_buff,temp);
			/*
			| flush the output buffer
			*/
		flush_out_buff();
		break;
	case(NONE):		/* Eliminate empty tags completely */
		status = strcmp(prev_tag,&temp[2]);
		if(!status){	/* tags match */
				/*
				| ignore the tag
				*/
			out_buff[0] = '\0';
		}else{
			strcat(out_buff,temp);
			flush_out_buff();
		}
		break;
	case(END):		/* trash closing tag of an empty tag set */
		status = strcmp(prev_tag,&temp[2]);
		if(!status){	/* tags match */
				/*
				| ignore the tag
				*/
			flush_out_buff();
		}else{
			strcat(out_buff,temp);
			flush_out_buff();
		}
		break;
	default:
		printf("Invalid empty_tag value!! %d\n",empty_tag);
		exit(1);
	}
	prev_tag[0] = '\0';
}
/***********************+-------------------+
|			| process_start_tag |
|			+-------------------+
| A good thing here is the tag is on a line by itself. The []
| must be replaced with <>.
*/
void process_start_tag()
{
	int  i = 1;
	int  j = 1;
	int  data_in_out_buff;
	char ch;
		/*
		| flush anything left over in the output buffer
		*/
	data_in_out_buff = strlen(out_buff);
	if(data_in_out_buff){
		flush_out_buff();
	}
		/*
		| Start building the new start tag in output buffer
		*/
	strcpy(out_buff,"<");
	ch        = in_buff[i++];
	do{
		out_buff[j++] = ch;
		ch           = in_buff[i++];
	}while(ch != END_CHAR);
	out_buff[j++] = '>';
	out_buff[j]   = '\0';
		/*
		| Output buffer is not flushed so the start tag
		| will end up on the same line as the data...
		*/
		/*
		| copy the start tag to save area to test for 
		| empty tag
		*/
	strcpy(prev_tag,&out_buff[1]);
}
/***********************+--------------+
|			| process_data |
|			+--------------+
| Date is bracketed with '|'. Programs ends up here when the
| start of data is detected. If data does not fit all on one
| line, the flush data will catch it. A global flag is set
| to tell flush_data() to continue writing.
*/
void process_data()
{
	int  i = 1;
	int  j = 0;
	int  len;
	char ch;
	char temp[256] = "";
		/*
		| There is data, so the prev tag is not empty
		*/
	prev_tag[0] = '\0';
		/*
		| Set flag in case there is a data continuation
		| condition.
		*/
	cont_flag = TRUE;
		/*
		| Start collecting temp string
		*/
	ch        = in_buff[i++];
	do{
		temp[j++] = ch;
		ch        = in_buff[i++];
//	}while((ch) && (ch != DATA_CHAR) && (ch != '\n'));
	}while((ch) && (ch != DATA_CHAR));
	if(ch < ' '){
		printf(">>>\n");
	}
	temp[j]   = '\0';
		/*
		| The temp string, which now contains the converted
		| end tag is tacked onto the output buffer...
		*/
	strcat(out_buff,temp);
		/*
		| Determine whether this is the last data record
		| or not. Clear flag if so...
		*/
	len = strlen(in_buff);
	if(in_buff[len-2]  == DATA_CHAR){
		cont_flag   = FALSE;
	}

}
/***********************+------------+
|			| flush_data |
|			+------------+
| If flush_data() is called and the continuation flag is not set,
| an error has occurred and must be handled. The continuation
| flag is cleared if the last character is '|';
*/
void flush_data()
{
	int  i;
	int  len;
	char ch;
		/*
		| flush the output buffer if it has data...
		*/
	if(strlen(out_buff)){
		flush_out_buff();
	}
		/*
		| Determine whether this is the last data record
		| or not. Clear flag if so...
		*/
	len  = strlen(in_buff);
	len -= 2;			/* back up before \n */
	if(in_buff[len] == DATA_CHAR){
		in_buff[len] = '\0';	/* lose the |\n  */
		cont_flag  = FALSE;
	}
	strcpy(out_buff,in_buff);
}
/***********************+--------------+
|			| process_rast |
|			+--------------+
| The basic plan here is to determine whether the line is:
|	1) a start tag
|	2) an end tag
|	3) A data record
|	4) a data continuation
| then process it. There is a fuction for each...
*/
void process_rast()
{
	if(in_buff[0] == START_CHAR){
		if(in_buff[1] == END_TAG){
			process_end_tag();
		}else{
			process_start_tag();
		}
	}else{
		if(in_buff[0] == DATA_CHAR){
			process_data();
		}else{
			flush_data();
		}
	}
}
/***********************+-------+
|			| usage |
|			+-------+
*/
void usage()
{
	printf("\n%s\n",ptitle);
	printf("%s\n",pdate);
	printf("\n");
	printf("Usage:   rast2sgm [-eE] file_in file_out\n\n");
	printf("Where:   -e Remove all empty tags\n");
	printf("         -E Remove end tag from empty tags\n");
	printf("         -w Don't wrap lines at #RE\n");
	printf("         file_in  is the input filename\n");
	printf("         file_out is the output filename\n\n");
	printf("The default is keeping empty tags\n");
	printf("RAST2SGM.EXE is placed in the Public Domain by the author, \n");
	printf("Norman E. Smith, 72745.1566@Compuserve.com\n\n");
	printf("RAST2SGM converts the output of RAST.EXE into SGML format\n");
	printf("RAST is part of the SGMLS package. Run SGMLS, then RAST,\n");
	printf("then RAST2SGM to expand a minimized SGML file to include\n");
	printf("end tags.\n\n");
}
/***********************+------+
|			| main |
|			+------+
*/
void main(int argc, char *argv[])
{
	int  i = 1;
	int  status;

	char *not_done;
	char flag;

	if(argc < 3){
		usage();
		exit(1);
	}
		/*
		| Check to see if there is a processing flag
		*/
	if(*argv[1] == '-'){
		flag = *(argv[1]+1);
		switch(flag){
		case 'e':
			empty_tag  = NONE;	/* ignore empty tags */
			break;
		case 'E':
			empty_tag  = END;
			break;
		case 'w':			/* Wrap long lines */
		case 'W':
			wrap_lines = FALSE;
			break;
		default:
			usage();
			exit(1);
		}
		i = 2;
	}
	strcpy(file_in,argv[i++]);
	strcpy(file_out,argv[i]);
	init_rast();
	FOREVER{
		not_done = read_next_line();
			/*
			| don't pass #RE lines through
			*/
		status = strcmp(in_buff,"#RE\n");
		if(!status){
			if(wrap_lines){
				in_buff[0] = '\0';
			}else{
				strcat(out_buff," ");
				continue;
			}
		}
		if(not_done){
			process_rast();
		}else{
			break;
		}
	}
	exit_rast();
}
